/*
**
**  Nintendo APU Emulator
**  Copyright 2000, Jason Cline (meridian@overclocked.org)
**  http://meridian.overclocked.org
**
**  Nintendo APU emulation routines for public use
**  If you make any improvements to this code send
**  them to me so they can be included in the
**  public release for everyone's benefit.
**  If you make any improvements to this code send
**
** History:
**
** 11.24.2001 Jason Cline:
**  - initial public release
*/

#include <memory.h>
#include <stdio.h>

//iData types
#define UINT32                    unsigned long int
#define UINT16                    unsigned short int
#define UINT8                     unsigned char
#define INT32                     long int
#define INT16                     short int
#define INT8                      char

//NTSC is 60 FPS
#define  FPS                      60

//macros
#define ZeroMemory(a, b) memset(a, 0, b); 

//necessary globals
static UINT32 iBufferLength=0;
static UINT32 iBufferInc=0;
static UINT8 *pCPUMemory=NULL;
static UINT32 iIRQ=0;

//square channels
struct{
  UINT8  iRegister[4];
  INT32  iLengthCounter;
  INT32  iFrequency;
  INT32  iLinearCounterPhase;
  UINT8  iDutyCyclePhase;
  INT32  iSweepPhase;
  INT32  iEnvelopePhase;
  UINT8  iEnvelopeVolume;
  UINT32 iEnabled;
}APUSquare[2];

//triangle channel
struct{
  UINT8  iRegister[4];
  INT32  iLinearLength;
  INT32  iLengthCounter;
  INT32  iWriteLatency;
  INT32  iLinearCounterPhase;
  UINT8  iStepGeneratorPhase;
  UINT32 iCounterStarted;
  UINT32 iEnabled;
}APUTriangle;

//noise channel
struct{
  UINT8  iRegister[4];
  UINT32 iCurrentPosition;
  INT32  iLengthCounter;
  INT32  iLinearCounterPhase;
  INT32  iEnvelopePhase;
  UINT8  iEnvelopeVolume;
  UINT8  iTwoCyclePhase;
  UINT32 iEnabled;
}APUNoise;

//pcm channel
struct{
  UINT8  iRegister[4];
  UINT32 iAddress;
  UINT32 iLengthCounter;
  UINT32 iBitPos;
  INT32  iLinearCounterPhase;
  UINT8  iCurrentByte;
  UINT32 iEnabled;
}APUDPCM;

//vblank Length table used for squares, triangle, noise
static const UINT8 iVBLength[]={
 0x05,0x7f,0x0a,0x01,0x14,0x02,0x28,0x03,0x50,0x04,0x1e,0x05,0x7,0x06,0x0e,0x07,
 0x06,0x08,0x0c,0x09,0x18,0x0a,0x30,0xb,0x60,0x0c,0x24,0x0d,0x8,0x0e,0x10,0x0f
};
//Frequency limit of square channels
static const INT32 iFrequencyLimit[]={
  0x3ff,0x555,0x666,0x71c,0x787,0x7c1,0x7e0,0x7f0
};
//table of noise Frequencies 
static const UINT32 iNoiseiFrequency[]={
  0x02,0x04,0x08,0x10,0x20,0x30,0x40,0x50,0x65,0x7f,0xbe,0xfe,0x17d,0x1fc,0x3f9,0x7f2
};
//dpcm transfer Frequencies
static const UINT32 iDPCMClocks[]={
  0xd60,0xbe0,0xaa0,0xa00,0x8f0,0x7f0,0x710,0x6b0,0x5f0,0x500,0x470,0x400,0x350,0x2a0,
  0x240,0x1b0
};
//ratios of pos/neg pulse for square waves
static const UINT32 iDutyLookUpTable[]={
  0x02, 0x04, 0x08, 0x0c
};

//HANDLES SQUARE CHANNEL A
INT8 mAPUSquareUpdateA()
{
static INT32 iEnvDelay;
static INT32 iSweepDelay;
static INT8  iOutput;

  //exit if the channel is off
  if(!APUSquare[0].iEnabled) return(0);
  //precompute some values
  iEnvDelay=((APUSquare[0].iRegister[0] & 0x0f) + 1) * iBufferLength;
  APUSquare[0].iEnvelopePhase-=240/60;
  //envelope decay unit
  while(APUSquare[0].iEnvelopePhase<0){
    APUSquare[0].iEnvelopePhase += iEnvDelay;
    if(APUSquare[0].iRegister[0] & 0x20) APUSquare[0].iEnvelopeVolume=(APUSquare[0].iEnvelopeVolume + 1) & 15;
    else if(APUSquare[0].iEnvelopeVolume < 15) APUSquare[0].iEnvelopeVolume++;
  }
  //length counter
  if((APUSquare[0].iRegister[0] & 0x20)==0) APUSquare[0].iLengthCounter-=(APUSquare[0].iLengthCounter > 0);
  if(APUSquare[0].iLengthCounter==0) return(0);
  //sweep unit
  if((APUSquare[0].iRegister[1] & 0x80) && (APUSquare[0].iRegister[1] & 7)){
    iSweepDelay=(((APUSquare[0].iRegister[1] >> 4) & 7) + 1) * iBufferLength;
    APUSquare[0].iSweepPhase-=120/60;
    while(APUSquare[0].iSweepPhase<0){
      APUSquare[0].iSweepPhase += iSweepDelay;
      if(APUSquare[0].iRegister[1] & 8) APUSquare[0].iFrequency -= ((APUSquare[0].iFrequency >> (APUSquare[0].iRegister[1] & 7)) + 1);
      else APUSquare[0].iFrequency += (APUSquare[0].iFrequency >> (APUSquare[0].iRegister[1] & 7));
    }
  }
  //linear iFrequency counter
  if(((APUSquare[0].iRegister[1] & 8)==0 && (APUSquare[0].iFrequency) > iFrequencyLimit[APUSquare[0].iRegister[1] & 7]) || (APUSquare[0].iFrequency) < 4) return(0);
  APUSquare[0].iLinearCounterPhase -= iBufferInc;
  while(APUSquare[0].iLinearCounterPhase < 0){
    APUSquare[0].iLinearCounterPhase += (APUSquare[0].iFrequency);
    APUSquare[0].iDutyCyclePhase=(APUSquare[0].iDutyCyclePhase + 1) & 0x0f;
  }
  //compute Output volume
  if(APUSquare[0].iRegister[0] & 0x10) iOutput=APUSquare[0].iRegister[0] & 0x0f;
  else iOutput=15 - APUSquare[0].iEnvelopeVolume;
  //check duty cycle to see if we are up or down
  if(APUSquare[0].iDutyCyclePhase < (iDutyLookUpTable[APUSquare[0].iRegister[0] >> 6])) iOutput=-iOutput;
  //we dont return the actual digital volume but instead 1.5 times the normal volume
  //because the RF connector on the actual NES modulated the iOutput
  return((iOutput * 3) / 2);
}

//HANDLES SQUARE CHANNEL B
INT8 mAPUSquareUpdateB()
{
static INT32 iEnvDelay;
static INT32 iSweepDelay;
static INT8  iOutput;

  //exit if the channel is off
  if(!APUSquare[1].iEnabled) return(0);
  //precompute some values
  iEnvDelay=((APUSquare[1].iRegister[0] & 0x0f) + 1) * iBufferLength;
  APUSquare[1].iEnvelopePhase-=240/60;
  //envelope decay unit
  while(APUSquare[1].iEnvelopePhase<0){
    APUSquare[1].iEnvelopePhase += iEnvDelay;
    if(APUSquare[1].iRegister[0] & 0x20) APUSquare[1].iEnvelopeVolume=(APUSquare[1].iEnvelopeVolume + 1) & 15;
    else if(APUSquare[1].iEnvelopeVolume < 15) APUSquare[1].iEnvelopeVolume++;
  }
  //length counter
  if((APUSquare[1].iRegister[0] & 0x20)==0) APUSquare[1].iLengthCounter-=(APUSquare[1].iLengthCounter > 0);
  if(APUSquare[1].iLengthCounter==0) return(0);
  //sweep unit
  if((APUSquare[1].iRegister[1] & 0x80) && (APUSquare[1].iRegister[1] & 7)){
    iSweepDelay=(((APUSquare[1].iRegister[1] >> 4) & 7) + 1) * iBufferLength;
    APUSquare[1].iSweepPhase-=120/60;
    while(APUSquare[1].iSweepPhase<0){
      APUSquare[1].iSweepPhase += iSweepDelay;
      if(APUSquare[1].iRegister[1] & 8) APUSquare[1].iFrequency -= (APUSquare[1].iFrequency >> (APUSquare[1].iRegister[1] & 7));
      else APUSquare[1].iFrequency += (APUSquare[1].iFrequency >> (APUSquare[1].iRegister[1] & 7));
    }
  }
  //linear iFrequency counter
  if(((APUSquare[1].iRegister[1] & 8)==0 && (APUSquare[1].iFrequency) > iFrequencyLimit[APUSquare[1].iRegister[1] & 7]) || (APUSquare[1].iFrequency) < 4) return(0);
  APUSquare[1].iLinearCounterPhase -= iBufferInc;
  while(APUSquare[1].iLinearCounterPhase < 0){
    APUSquare[1].iLinearCounterPhase += (APUSquare[1].iFrequency);
    APUSquare[1].iDutyCyclePhase=(APUSquare[1].iDutyCyclePhase + 1) & 0x0f;
  }
  //compute Output volume
  if(APUSquare[1].iRegister[0] & 0x10) iOutput=APUSquare[1].iRegister[0] & 0x0f;
  else iOutput=15 - APUSquare[1].iEnvelopeVolume;
  //check duty cycle to see if we are up or down
  if(APUSquare[1].iDutyCyclePhase < (iDutyLookUpTable[APUSquare[1].iRegister[0] >> 6])) iOutput=-iOutput;
  //we dont return the actual digital volume but instead 1.5 times the normal volume
  //because the RF connector on the actual NES amplified the iOutput on this channel somehow
  return((iOutput * 3) / 2);
}

//HANDLES THE TRIANGLE CHANNEL
INT8 mAPUTriangleUpdate()
{
static INT32 iFrequency;
static INT8  iOutput;

  //exit if the channel is off
  if(!APUTriangle.iEnabled) return(0);
  //start the counter if it isn't started
  if(!APUTriangle.iCounterStarted && (APUTriangle.iRegister[0] & 0x80)==0){
    if(APUTriangle.iWriteLatency) APUTriangle.iWriteLatency--;
    if(APUTriangle.iWriteLatency==0) APUTriangle.iCounterStarted=1;
  }
  //length counter
  if(APUTriangle.iCounterStarted){
    if(APUTriangle.iLinearLength > 0) APUTriangle.iLinearLength--;
    if(APUTriangle.iLengthCounter && (APUTriangle.iRegister[0] & 0x80)==0) APUTriangle.iLengthCounter--;
    if(APUTriangle.iLengthCounter==0) return(0);
  }
  if(APUTriangle.iLinearLength==0) return(0);
  //computer the iFrequency
  iFrequency=(((APUTriangle.iRegister[3] & 7) << 8) + APUTriangle.iRegister[2]) + 1;
  if(iFrequency < 4) return(0);
  //linear iFrequency counter
  APUTriangle.iLinearCounterPhase -= iBufferInc;
  while(APUTriangle.iLinearCounterPhase < 0){
    APUTriangle.iLinearCounterPhase += iFrequency;
    APUTriangle.iStepGeneratorPhase=(APUTriangle.iStepGeneratorPhase + 1) & 0x1f;
    iOutput=(APUTriangle.iStepGeneratorPhase & 7) << 1;
    if(APUTriangle.iStepGeneratorPhase & 8) iOutput=0x10 - iOutput;
    if(APUTriangle.iStepGeneratorPhase & 0x10) iOutput=-iOutput;
  }
  //return the actual digital volume for this channel
  return(iOutput);
}

//HANDLES THE NOISE CHANNEL
INT8 mAPUNoiseUpdate()
{
static INT32  iFrequency, iEnvDelay;
static UINT32 I;
static INT8   iOutput;

  //exit if the channel is off
  if(!APUNoise.iEnabled) return(0);
  //envelope decay unit
  iEnvDelay=((APUNoise.iRegister[0] & 0x0f) + 1) * iBufferLength;
  APUNoise.iEnvelopePhase -= 240/60;
  while (APUNoise.iEnvelopePhase < 0){
    APUNoise.iEnvelopePhase += iEnvDelay;
    if(APUNoise.iRegister[0] & 0x20) APUNoise.iEnvelopeVolume=(APUNoise.iEnvelopeVolume + 1) & 15;
    else if(APUNoise.iEnvelopeVolume < 15) APUNoise.iEnvelopeVolume++;
  }
  //length counter
  if((APUNoise.iRegister[0] & 0x20)==0) APUNoise.iLengthCounter-=(APUNoise.iLengthCounter > 0);
  if(APUNoise.iLengthCounter==0) return(0);
  //linear iFrequency counter
  iFrequency=iNoiseiFrequency[APUNoise.iRegister[2] & 0x0f];
  APUNoise.iLinearCounterPhase -= iBufferInc;
  while(APUNoise.iLinearCounterPhase < 0){
    APUNoise.iLinearCounterPhase += iFrequency;
    //half the Output iFrequency for shifting
    APUNoise.iTwoCyclePhase=(APUNoise.iTwoCyclePhase + 1) & 1;
    if(APUNoise.iTwoCyclePhase){
      //short noise
      if(APUNoise.iRegister[2] & 0x80){
        I=((APUNoise.iCurrentPosition >> 8) & 1) ^ ((APUNoise.iCurrentPosition >> 14) & 1);
        APUNoise.iCurrentPosition=(APUNoise.iCurrentPosition << 1) | I;
        APUNoise.iCurrentPosition &= 0x7fff;
      }
      //long noise
      else{
        I=((APUNoise.iCurrentPosition >> 13) & 1) ^ ((APUNoise.iCurrentPosition >> 14) & 1);
        APUNoise.iCurrentPosition=(APUNoise.iCurrentPosition << 1) | I;
        APUNoise.iCurrentPosition &= 0x7fff;
      }
    }
  }
  //compute iOutput volume
  if(APUNoise.iRegister[0] & 0x10) iOutput=APUNoise.iRegister[0] & 0x0f;
  else iOutput=15 - APUNoise.iEnvelopeVolume;
  if(!((APUNoise.iCurrentPosition >> 14) & 1)) iOutput=-iOutput;
  //we dont return the actual digital volume but instead 0.5 times the normal volume
  //because the RF connector on the actual NES filtered out some of the noise from the iOutput
  return(iOutput / 2);
}

//HANDLES THE PCM CHANNEL
INT8 mAPUDPCMUpdate()
{
static UINT32 iFrequency;

  if(APUDPCM.iEnabled && APUDPCM.iLengthCounter){
    //computer the iFrequency
    iFrequency=iDPCMClocks[APUDPCM.iRegister[0] & 0x0f] >> 3;
    //linear iFrequency counter
    APUDPCM.iLinearCounterPhase-=iBufferInc;
    while(APUDPCM.iLinearCounterPhase < 0){
      APUDPCM.iLinearCounterPhase+=iFrequency;
      APUDPCM.iCurrentByte=pCPUMemory[APUDPCM.iAddress];
      //update the dpmc iValue
      if(APUDPCM.iCurrentByte & (1 << APUDPCM.iBitPos)) APUDPCM.iRegister[1]+=(APUDPCM.iRegister[1] < 63);
      else APUDPCM.iRegister[1]-=(APUDPCM.iRegister[1] > 0);
      //check for dma access
      if(APUDPCM.iBitPos==7){
        //check for end of playback
        if(APUDPCM.iLengthCounter==0){
          //looping iEnabled
          if(APUDPCM.iRegister[0] & 0x40){
            APUDPCM.iAddress=0xc000 + (APUDPCM.iRegister[2] << 6);
            APUDPCM.iLengthCounter=APUDPCM.iRegister[3] << 4;
            APUDPCM.iBitPos=0;
            iIRQ=0;
          }
          //looping disabled
          else{
            APUDPCM.iEnabled=0;
            iIRQ=1;
            //we dont return the actual digital volume but instead 2.0 times the normal volume
            //because the RF connector on the actual NES amplified the iOutput on this channel somehow
            return((APUDPCM.iRegister[1] - 32) * 2);
          }
        }
        //next byte
        else{
          APUDPCM.iLengthCounter--;
          APUDPCM.iBitPos=0;
          if(APUDPCM.iAddress==0xffff) APUDPCM.iAddress=0x8000;
          else APUDPCM.iAddress++;
        }
      }
      //goto next bit
      else APUDPCM.iBitPos++;
    }
    //we dont return the actual digital volume but instead 2.0 times the normal volume
    //because the RF connector on the actual NES amplified the iOutput on this channel somehow
    return((APUDPCM.iRegister[1] - 32) * 2);
  }
  return(0);
}

//RETURN IRQ STATUS OF THE PCM CHANNEL
UINT32 mIRQStatus()
{
UINT32 tIRQ;

  tIRQ=iIRQ;
  //clear IRQ status after read
  iIRQ=0;
  return(tIRQ);
}

//HANDLES A READ FROM THE APU
UINT8 mAPURead(UINT32 iAddress)
{
UINT8 iValue;

  switch(iAddress){
   case 0x4015:
     iValue=0;
     if(APUSquare[0].iEnabled && APUSquare[0].iLengthCounter) iValue |= 0x01;
     if(APUSquare[1].iEnabled && APUSquare[1].iLengthCounter) iValue |= 0x02;
     if(APUTriangle.iEnabled && APUTriangle.iLengthCounter) iValue |= 0x04;
     if(APUNoise.iEnabled && APUNoise.iLengthCounter) iValue |= 0x08;
     if(APUDPCM.iEnabled && APUDPCM.iLengthCounter) iValue |= 0x10;
     iValue |= 0x40;
     if(iIRQ) iValue |= 0x80;
     return(iValue);
  }
  return(0x40);
}

//HANDLES A WRITE TO THE APU
void mAPUWrite(UINT32 iAddress, UINT8 iData)
{
UINT32 Chan;

  switch(iAddress){
    //Square Channels
    case 0x4000:
    case 0x4004:
      Chan=(iAddress & 0x4) ? 1 : 0;
      APUSquare[Chan].iRegister[0]=iData;
      break;
    case 0x4001:
    case 0x4005:
      Chan=(iAddress & 0x4) ? 1 : 0;
      APUSquare[Chan].iRegister[1]=iData;
      break;
    case 0x4002:
    case 0x4006:
      Chan=(iAddress & 0x4) ? 1 : 0;
      APUSquare[Chan].iRegister[2]=iData;
      if(APUSquare[Chan].iEnabled) APUSquare[Chan].iFrequency=((((APUSquare[Chan].iRegister[3] & 7) << 8) + iData) + 1);
      break;
    case 0x4003:
    case 0x4007:
      Chan=(iAddress & 0x4) ? 1 : 0;
      APUSquare[Chan].iRegister[3]=iData;
      if(APUSquare[Chan].iEnabled){
        APUSquare[Chan].iLengthCounter=iVBLength[iData >> 3]*iBufferLength;
        APUSquare[Chan].iEnvelopeVolume=0;
        APUSquare[Chan].iFrequency=((((iData & 7) << 8) + APUSquare[Chan].iRegister[2]) + 1);
      }
      break;
    //Triangle Channel
    case 0x4008:
      APUTriangle.iRegister[0]=iData;
      if(APUTriangle.iEnabled && !APUTriangle.iCounterStarted) APUTriangle.iLinearLength=((iData& 0x7f) * iBufferLength) >> 2;
      break;
    case 0x4009:
      //unused
      APUTriangle.iRegister[1]=iData;
      break;
    case 0x400a:
      APUTriangle.iRegister[2]=iData;
      break;
    case 0x400b:
      APUTriangle.iRegister[3]=iData;
      APUTriangle.iWriteLatency=3;
      if(APUTriangle.iEnabled){
        APUTriangle.iCounterStarted=0;
        APUTriangle.iLengthCounter=iVBLength[iData >> 3] * iBufferLength;
        APUTriangle.iLinearLength=((APUTriangle.iRegister[0] & 0x7f) * iBufferLength) >> 2;
      }
      break;
    //Noise Channel
    case 0x400c:
      APUNoise.iRegister[0]=iData;
      break;
    case 0x400d:
      //unused
      APUNoise.iRegister[1]=iData;
      break;
    case 0x400e:
      APUNoise.iRegister[2]=iData;
      break;
    case 0x400f:
      APUNoise.iRegister[3]=iData;
      if(APUNoise.iEnabled){
        APUNoise.iLengthCounter=iVBLength[iData >> 3] * iBufferLength;
        APUNoise.iEnvelopeVolume=0;
      }
      break;
    //PCM Channel
    case 0x4010:
      APUDPCM.iRegister[0]=iData;
      if((iData & 0x80)==0) iIRQ=0;
      break;
    case 0x4011:
      APUDPCM.iRegister[1]=(iData & 0x7f) >> 1;
      break;
    case 0x4012:
      APUDPCM.iRegister[2]=iData;
      APUDPCM.iAddress=0xc000 + (APUDPCM.iRegister[2] << 6);
      break;
    case 0x4013:
      APUDPCM.iRegister[3]=iData;
      APUDPCM.iLengthCounter=APUDPCM.iRegister[3] << 4;
      APUDPCM.iBitPos=0;
      break;
    //Status Mask
    case 0x4015:
      if(iData & 0x01) APUSquare[0].iEnabled=1;
      else{
        APUSquare[0].iEnabled=0;
        APUSquare[0].iLengthCounter=0;
      }
      if(iData & 0x02) APUSquare[1].iEnabled=1;
      else{
        APUSquare[1].iEnabled=0;
        APUSquare[1].iLengthCounter=0;
      }
      if(iData & 0x04) APUTriangle.iEnabled=1;
      else{
        APUTriangle.iEnabled=0;
        APUTriangle.iLengthCounter=0;
        APUTriangle.iLinearLength=0;
        APUTriangle.iCounterStarted=0;
        APUTriangle.iWriteLatency=0;
      }
      if(iData & 0x08) APUNoise.iEnabled=1;
      else{
        APUNoise.iEnabled=0;
        APUNoise.iLengthCounter=0;
      }
      if(iData & 0x10){
        if(!APUDPCM.iEnabled){
          APUDPCM.iAddress=0xc000 + (APUDPCM.iRegister[2] << 6);
          APUDPCM.iLengthCounter=APUDPCM.iRegister[3] << 4;
          APUDPCM.iBitPos=0;
        }
        APUDPCM.iEnabled=1;
      }
      else APUDPCM.iEnabled=0;
      iIRQ=0;
      break;
    default:
      break;
  }
}

//RESET THE APU
void mAPUReset(UINT32 Samplerate, UINT8 *pCPU)
{
  pCPUMemory=pCPU;
  iBufferLength=Samplerate / FPS;
  iBufferInc=1790000 / Samplerate;
  ZeroMemory(&APUSquare[0], sizeof(APUSquare[0]));
  ZeroMemory(&APUSquare[1], sizeof(APUSquare[1]));
  ZeroMemory(&APUTriangle, sizeof(APUTriangle));
  ZeroMemory(&APUNoise, sizeof(APUNoise));
  ZeroMemory(&APUDPCM, sizeof(APUDPCM));
  APUSquare[0].iEnabled=1;
  APUSquare[1].iEnabled=1;
  APUTriangle.iEnabled=1;
  APUNoise.iEnabled=1;
  APUNoise.iCurrentPosition=1;
  APUDPCM.iEnabled=0;
  APUDPCM.iRegister[1]=0;
}