/*
**
**  Nintendo MMC5 Emulator
**  Copyright 2000, Jason Cline (meridian@overclocked.org)
**  http://meridian.overclocked.org
**
**  Nintendo MMC5 emulation routines for public use
**  If you make any improvements to this code send
**  them to me so they can be included in the
**  public release for everyone's benefit.
**  If you make any improvements to this code send
**
** History:
**
** 11.24.2001 Jason Cline:
**  - initial public release
*/


#include <memory.h>
#include <stdio.h>

//Data types
#define UINT32    unsigned long int
#define UINT16    unsigned short int
#define UINT8     unsigned char
#define INT32     long int
#define INT16     short int
#define INT8      char

//NTSC is 60 FPS
#define  FPS                      60

//macros
#define ZeroMemory(a, b) memset(a, 0, b); 

//necessary globals
static UINT32 iBufferLength=0;
static UINT32 iBufferInc=0;

//square channels
struct{
  UINT8  iRegister[4];
  INT32  iLengthCounter;
  INT32  iFrequency;
  INT32  iLinearCounterPhase;
  UINT8  iDutyCyclePhase;
  INT32  iSweepPhase;
  INT32  iEnvelopePhase;
  UINT8  iEnvelopeVolume;
  UINT32 iEnabled;
}MMC5Square[2];

struct{
  UINT8  iRegister[2];
  UINT32 iEnabled;
}MMC5Wave;

//vblank Length table used for squares, triangle, noise
static const UINT8 iVBLength[]={
 0x05,0x7f,0x0a,0x01,0x14,0x02,0x28,0x03,0x50,0x04,0x1e,0x05,0x7,0x06,0x0e,0x07,
 0x06,0x08,0x0c,0x09,0x18,0x0a,0x30,0xb,0x60,0x0c,0x24,0x0d,0x8,0x0e,0x10,0x0f
};
//Frequency limit of square channels
static const INT32 iFrequencyLimit[]={
  0x3ff,0x555,0x666,0x71c,0x787,0x7c1,0x7e0,0x7f0
};
//ratios of pos/neg pulse for square waves
static const UINT32 iDutyLookUpTable[]={
  0x02, 0x04, 0x08, 0x0c
};

UINT8 iMulA, iMulB;

//HANDLES SQUARE CHANNEL A
INT8 mMMC5SquareUpdateA()
{
static INT32 iEnvDelay;
static INT32 iSweepDelay;
static INT8  iOutput;

  //exit if the channel is off
  if(!MMC5Square[0].iEnabled) return(0);
  //precompute some values
  iEnvDelay=((MMC5Square[0].iRegister[0] & 0x0f) + 1) * iBufferLength;
  MMC5Square[0].iEnvelopePhase-=240/60;
  //envelope decay unit
  while(MMC5Square[0].iEnvelopePhase<0){
    MMC5Square[0].iEnvelopePhase += iEnvDelay;
    if(MMC5Square[0].iRegister[0] & 0x20) MMC5Square[0].iEnvelopeVolume=(MMC5Square[0].iEnvelopeVolume + 1) & 15;
    else if(MMC5Square[0].iEnvelopeVolume < 15) MMC5Square[0].iEnvelopeVolume++;
  }
  //length counter
  if((MMC5Square[0].iRegister[0] & 0x20)==0) MMC5Square[0].iLengthCounter-=(MMC5Square[0].iLengthCounter > 0);
  if(MMC5Square[0].iLengthCounter==0) return(0);
  //sweep unit
  if((MMC5Square[0].iRegister[1] & 0x80) && (MMC5Square[0].iRegister[1] & 7)){
    iSweepDelay=(((MMC5Square[0].iRegister[1] >> 4) & 7) + 1) * iBufferLength;
    MMC5Square[0].iSweepPhase-=120/60;
    while(MMC5Square[0].iSweepPhase<0){
      MMC5Square[0].iSweepPhase += iSweepDelay;
      if(MMC5Square[0].iRegister[1] & 8) MMC5Square[0].iFrequency -= ((MMC5Square[0].iFrequency >> (MMC5Square[0].iRegister[1] & 7)) + 1);
      else MMC5Square[0].iFrequency += (MMC5Square[0].iFrequency >> (MMC5Square[0].iRegister[1] & 7));
    }
  }
  //linear iFrequency counter
  if(((MMC5Square[0].iRegister[1] & 8)==0 && (MMC5Square[0].iFrequency) > iFrequencyLimit[MMC5Square[0].iRegister[1] & 7]) || (MMC5Square[0].iFrequency) < 4) return(0);
  MMC5Square[0].iLinearCounterPhase -= iBufferInc;
  while(MMC5Square[0].iLinearCounterPhase < 0){
    MMC5Square[0].iLinearCounterPhase += (MMC5Square[0].iFrequency);
    MMC5Square[0].iDutyCyclePhase=(MMC5Square[0].iDutyCyclePhase + 1) & 0x0f;
  }
  //compute iOutput volume
  if(MMC5Square[0].iRegister[0] & 0x10) iOutput=MMC5Square[0].iRegister[0] & 0x0f;
  else iOutput=15 - MMC5Square[0].iEnvelopeVolume;
  //check duty cycle to see if we are up or down
  if(MMC5Square[0].iDutyCyclePhase < (iDutyLookUpTable[MMC5Square[0].iRegister[0] >> 6])) iOutput=-iOutput;
  //we dont return the actual digital volume but instead 1.5 times the normal volume
  //because the RF connector on the actual NES modulated the iOutput
  return((iOutput * 3) / 2);
}

//HANDLES SQUARE CHANNEL B
INT8 mMMC5SquareUpdateB()
{
static INT32 iEnvDelay;
static INT32 iSweepDelay;
static INT8  iOutput;

  //exit if the channel is off
  if(!MMC5Square[1].iEnabled) return(0);
  //precompute some values
  iEnvDelay=((MMC5Square[1].iRegister[0] & 0x0f) + 1) * iBufferLength;
  MMC5Square[1].iEnvelopePhase-=240/60;
  //envelope decay unit
  while(MMC5Square[1].iEnvelopePhase<0){
    MMC5Square[1].iEnvelopePhase += iEnvDelay;
    if(MMC5Square[1].iRegister[0] & 0x20) MMC5Square[1].iEnvelopeVolume=(MMC5Square[1].iEnvelopeVolume + 1) & 15;
    else if(MMC5Square[1].iEnvelopeVolume < 15) MMC5Square[1].iEnvelopeVolume++;
  }
  //length counter
  if((MMC5Square[1].iRegister[0] & 0x20)==0) MMC5Square[1].iLengthCounter-=(MMC5Square[1].iLengthCounter > 0);
  if(MMC5Square[1].iLengthCounter==0) return(0);
  //sweep unit
  if((MMC5Square[1].iRegister[1] & 0x80) && (MMC5Square[1].iRegister[1] & 7)){
    iSweepDelay=(((MMC5Square[1].iRegister[1] >> 4) & 7) + 1) * iBufferLength;
    MMC5Square[1].iSweepPhase-=120/60;
    while(MMC5Square[1].iSweepPhase<0){
      MMC5Square[1].iSweepPhase += iSweepDelay;
      if(MMC5Square[1].iRegister[1] & 8) MMC5Square[1].iFrequency -= (MMC5Square[1].iFrequency >> (MMC5Square[1].iRegister[1] & 7));
      else MMC5Square[1].iFrequency += (MMC5Square[1].iFrequency >> (MMC5Square[1].iRegister[1] & 7));
    }
  }
  //linear iFrequency counter
  if(((MMC5Square[1].iRegister[1] & 8)==0 && (MMC5Square[1].iFrequency) > iFrequencyLimit[MMC5Square[1].iRegister[1] & 7]) || (MMC5Square[1].iFrequency) < 4) return(0);
  MMC5Square[1].iLinearCounterPhase -= iBufferInc;
  while(MMC5Square[1].iLinearCounterPhase < 0){
    MMC5Square[1].iLinearCounterPhase += (MMC5Square[1].iFrequency);
    MMC5Square[1].iDutyCyclePhase=(MMC5Square[1].iDutyCyclePhase + 1) & 0x0f;
  }
  //compute iOutput volume
  if(MMC5Square[1].iRegister[0] & 0x10) iOutput=MMC5Square[1].iRegister[0] & 0x0f;
  else iOutput=15 - MMC5Square[1].iEnvelopeVolume;
  //check duty cycle to see if we are up or down
  if(MMC5Square[1].iDutyCyclePhase < (iDutyLookUpTable[MMC5Square[1].iRegister[0] >> 6])) iOutput=-iOutput;
  //we dont return the actual digital volume but instead 1.5 times the normal volume
  //because the RF connector on the actual NES amplified the iOutput on this channel somehow
  return((iOutput * 3) / 2);
}

//HANDLES PCM CHANNEL
INT8 mMMC5PCMUpdate()
{
  if(!MMC5Wave.iEnabled) return(0);
  else return(MMC5Wave.iRegister[1] ^ 0x80);
}

//HANDLES A READ FROM THE MMC5
UINT8 mMMC5Read(UINT32 iAddress)
{
UINT32 iMul=iMulA * iMulB;

  switch(iAddress){
    case 0x5205:
      return(iMul & 0xff);
      break;
    case 0x5206:
      return((iMul >> 8) & 0xff);
      break;
  }
  return(iMulA * iMulB);
}

//HANDLES A WRITE TO THE MMC5
void mMMC5Write(UINT32 iAddress, UINT8 iData)
{
UINT32 Chan;

  Chan=(iAddress & 0x4) ? 1 : 0;
  switch(iAddress){
    //Square Channels
    case 0x5000:
    case 0x5004:
      MMC5Square[Chan].iRegister[0]=iData;
      break;
    case 0x5001:
    case 0x5005:
      MMC5Square[Chan].iRegister[1]=iData;
      break;
    case 0x5002:
    case 0x5006:
      MMC5Square[Chan].iRegister[2]=iData;
      if(MMC5Square[Chan].iEnabled) MMC5Square[Chan].iFrequency=((((MMC5Square[Chan].iRegister[3] & 7) << 8) + iData) + 1);
      break;
    case 0x5003:
    case 0x5007:
      MMC5Square[Chan].iRegister[3]=iData;
      if(MMC5Square[Chan].iEnabled){
        MMC5Square[Chan].iLengthCounter=iVBLength[iData >> 3]*iBufferLength;
        MMC5Square[Chan].iEnvelopeVolume=0;
        MMC5Square[Chan].iFrequency=((((iData & 7) << 8) + MMC5Square[Chan].iRegister[2]) + 1);
      }
      break;
    //Wave channel
    case 0x5010:
      MMC5Wave.iRegister[0]=iData;
      if(iData & 0x01) MMC5Wave.iEnabled=1;
      else MMC5Wave.iEnabled=0;
      break;
    case 0x5011:
      MMC5Wave.iRegister[1]=iData;
      break;
    //Status register
    case 0x5015:
      if(iData & 0x01) MMC5Square[0].iEnabled=1;
      else MMC5Square[0].iEnabled=0;
      if(iData & 0x02) MMC5Square[1].iEnabled=1;
      else MMC5Square[1].iEnabled=0;
      break;
    case 0x5205:
      iMulA=iData;
      break;
    case 0x5206:
      iMulB=iData;
      break;
  }
}

void mMMC5Reset(UINT32 Samplerate)
{
  iBufferLength=Samplerate / FPS;
  iBufferInc=1790000 / Samplerate;
  ZeroMemory(&MMC5Square[0], sizeof(MMC5Square[0]));
  ZeroMemory(&MMC5Square[1], sizeof(MMC5Square[1]));
  MMC5Square[0].iEnabled=0;
  MMC5Square[1].iEnabled=0;
  MMC5Wave.iEnabled=0;
}