/*
**
**  Konami VRC6 Emulator
**  Copyright 2000, Jason Cline (meridian@overclocked.org)
**  http://meridian.overclocked.org
**
**  Konami VRC6 emulation routines for public use
**  If you make any improvements to this code send
**  them to me so they can be included in the
**  public release for everyone's benefit.
**  If you make any improvements to this code send
**
** History:
**
** 11.24.2001 Jason Cline:
**  - initial public release
*/


#include <memory.h>
#include <stdio.h>

//Data types
#define UINT32    unsigned long int
#define UINT16    unsigned short int
#define UINT8     unsigned char
#define INT32     long int
#define INT16     short int
#define INT8      char

//NTSC is 60 FPS
#define  FPS                      60

//macros
#define ZeroMemory(a, b) memset(a, 0, b); 

//necessary globals
static UINT32 BufferLength=0;
static UINT32 BufferInc=0;

//square channels
struct{
  UINT8  iRegister[3];
  UINT32 iFrequency;
  INT32  iLinearCounterPhase;
  UINT8  iVolume;
  UINT8  iDutySet;
  UINT8  iDutyCyclePhase;
  UINT32 iEnabled;
}VRC6Square[2];

//sawtooth channel
struct{
  UINT8  iRegister[3];
  UINT32 iFrequency;
  INT32  iLinearCounterPhase;
  UINT8  iAccumulatorInput;
  UINT8  iAccumulator;
  UINT32 iEnabled;
}VRC6Sawtooth;

//ratios of pos/neg pulse for square waves
static const UINT32 iVRC6DutyLookUpTable[]={
  0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08
};

//HANDLES SQUARE CHANNEL A
INT8 mVRC6SquareUpdateA()
{
static INT8 iOutput;

  //exit if the channel is off
  if(!VRC6Square[0].iEnabled) return(0);
  //digital mode
  if(VRC6Square[0].iRegister[0] & 0x80) iOutput=VRC6Square[0].iVolume;
  //linear frequency counter
  else{
    VRC6Square[0].iLinearCounterPhase -= BufferInc;
    while(VRC6Square[0].iLinearCounterPhase < 0){
      VRC6Square[0].iLinearCounterPhase += VRC6Square[0].iFrequency;
      VRC6Square[0].iDutyCyclePhase=(VRC6Square[0].iDutyCyclePhase + 1) & 0x0f;
    }
    iOutput=VRC6Square[0].iVolume;
    //check duty cycle to see if we are up or down
    if(VRC6Square[0].iDutyCyclePhase < (iVRC6DutyLookUpTable[VRC6Square[0].iDutySet])) iOutput=-iOutput;
  }
  //we dont return the actual digital volume but instead 1.5 times the normal volume
  //because the RF connector on the actual NES modulated the iOutput
  return((iOutput * 3) / 2);
}

//HANDLES SQUARE CHANNEL B
INT8 mVRC6SquareUpdateB()
{
static INT8 iOutput=0;

  //exit if the channel is off
  if(!VRC6Square[1].iEnabled) return(0);
  //digital mode
  if(VRC6Square[1].iRegister[0] & 0x80) iOutput=VRC6Square[1].iVolume;
  //linear frequency counter
  else{
    VRC6Square[1].iLinearCounterPhase -= BufferInc;
    while(VRC6Square[1].iLinearCounterPhase < 0){
      VRC6Square[1].iLinearCounterPhase += VRC6Square[1].iFrequency;
      VRC6Square[1].iDutyCyclePhase=(VRC6Square[1].iDutyCyclePhase + 1) & 0x0f;
    }
    iOutput=VRC6Square[1].iVolume;
    //check duty cycle to see if we are up or down
    if(VRC6Square[1].iDutyCyclePhase < (iVRC6DutyLookUpTable[VRC6Square[1].iDutySet])) iOutput=-iOutput;
  }
  //we dont return the actual digital volume but instead 1.5 times the normal volume
  //because the RF connector on the actual NES modulated the iOutput
  return((iOutput * 3) / 2);
}

//HANDLES SAWTOOTH CHANNEL
INT8 mVRC6SawtoothUpdate()
{
static INT8 iOutput;

  //exit if the channel is off
  if(!VRC6Sawtooth.iEnabled) return(0);
  //linear frequency counter
  VRC6Sawtooth.iLinearCounterPhase -= BufferInc;
  while(VRC6Sawtooth.iLinearCounterPhase < 0){
    VRC6Sawtooth.iLinearCounterPhase += VRC6Sawtooth.iFrequency;
    iOutput += VRC6Sawtooth.iAccumulatorInput;
    VRC6Sawtooth.iAccumulator++;
    if(VRC6Sawtooth.iAccumulator==7){
      VRC6Sawtooth.iAccumulator=0;
      iOutput=0;
    }
  }
  //only the top 5 bits are used
  iOutput >>= 3;
  //return the actual digital volume for this channel
  return(iOutput);
}

//HANDLES A WRITE TO THE VRC6
void mVRC6Write(UINT32 Address, UINT8 Data)
{
  switch(Address){
    //Square Channels
    case 0x9000:
      VRC6Square[0].iRegister[0]=Data;
      VRC6Square[0].iVolume=Data & 0xf;
      VRC6Square[0].iDutySet=(Data >> 4) & 0x7;
      break;
    case 0x9001:
      VRC6Square[0].iRegister[1]=Data;
      VRC6Square[0].iFrequency=((VRC6Square[2].iRegister[2] & 0xf) << 8) | VRC6Square[0].iRegister[1];
      break;
    case 0x9002:
      VRC6Square[0].iRegister[2]=Data;
      VRC6Square[0].iFrequency=((VRC6Square[0].iRegister[2] & 0xf) << 8) | VRC6Square[0].iRegister[1];
      VRC6Square[0].iEnabled=Data & 0x80;
      break;
    case 0xA000:
      VRC6Square[1].iRegister[0]=Data;
      VRC6Square[1].iVolume=Data & 0xf;
      VRC6Square[1].iDutySet=(Data >> 4) & 0x7;
      break;
    case 0xA001:
      VRC6Square[1].iRegister[1]=Data;
      VRC6Square[1].iFrequency=((VRC6Square[1].iRegister[2] & 0xf) << 8) | VRC6Square[1].iRegister[1];
      break;
    case 0xA002:
      VRC6Square[1].iRegister[2]=Data;
      VRC6Square[1].iFrequency=((VRC6Square[1].iRegister[2] & 0xf) << 8) | VRC6Square[1].iRegister[1];
      VRC6Square[1].iEnabled=Data & 0x80;
      break;
    //Sawtooth Channel
    case 0xB000:
      VRC6Sawtooth.iRegister[0]=Data;
      VRC6Sawtooth.iAccumulatorInput=Data & 0x3f;
      break;
    case 0xB001:
      VRC6Sawtooth.iRegister[1]=Data;
      VRC6Sawtooth.iFrequency=((VRC6Sawtooth.iRegister[2] & 0xf) << 8) | VRC6Sawtooth.iRegister[1];
      break;
    case 0xB002:
      VRC6Sawtooth.iRegister[2]=Data;
      VRC6Sawtooth.iFrequency=((VRC6Sawtooth.iRegister[2] & 0xf) << 8) | VRC6Sawtooth.iRegister[1];
      VRC6Sawtooth.iEnabled=Data & 0x80;
      break;
    default:
      break;
  }
}

//RESET THE VRC6
void mVRC6Reset(UINT32 Samplerate)
{
  BufferLength=Samplerate / FPS;
  BufferInc=1790000 / Samplerate;
  ZeroMemory(&VRC6Square[0], sizeof(VRC6Square[0]));
  ZeroMemory(&VRC6Square[1], sizeof(VRC6Square[1]));
  ZeroMemory(&VRC6Sawtooth, sizeof(VRC6Sawtooth));
  VRC6Square[0].iEnabled=0;
  VRC6Square[1].iEnabled=0;
  VRC6Sawtooth.iEnabled=0;
}